from typing import Any, List, Literal

from ....utils.class_registry import register_class


from ...modifiable_values import DamageValue
from ...event import (
    CharactorReviveEventArguments, CreateObjectEventArguments, 
    GameStartEventArguments, RoundEndEventArguments, SkillEndEventArguments
)
from ...action import Actions, CreateObjectAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, PassiveSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillBase, SkillTalent
)


# Skills


class AttachRiptideWhenSkillEnd(SkillBase):
    version: str

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When skill end, if is this skill, and target
        is alive, attach riptide to target.
        """
        if event.action.position.id != self.id:
            # not self skill
            return []
        target = match.player_tables[1 - self.position.player_idx].charactors[
            event.action.target_position.charactor_idx]
        if target.is_defeated:
            # target is defeated
            return []
        # attach riptide to target
        return [
            CreateObjectAction(
                object_name = 'Riptide',
                object_position = event.action.target_position.set_area(
                    ObjectPositionType.CHARACTOR_STATUS),
                object_arguments = { 'version': self.version }
            )
        ]


class FoulLegacyRagingTide(ElementalSkillBase, AttachRiptideWhenSkillEnd):
    name: Literal['Foul Legacy: Raging Tide'] = 'Foul Legacy: Raging Tide'
    version: Literal['4.1'] = '4.1' 
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object then attack
        """
        return [
            self.create_charactor_status(
                'Melee Stance',
                { 'version': self.version }
            ),
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.charge_self(1),
        ]


class HavocObliteration(ElementalBurstBase, AttachRiptideWhenSkillEnd):
    name: Literal['Havoc: Obliteration'] = 'Havoc: Obliteration'
    version: Literal['4.1'] = '4.1' 
    damage: int = 7
    ranged_damage: int = 5
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 3,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        ranged_found: bool = False
        melee_found: bool = False
        for status in charactor.status:
            if status.name == 'Ranged Stance':
                ranged_found = True
            elif status.name == 'Melee Stance':  # pragma: no branch
                melee_found = True
        if ranged_found and melee_found:
            raise AssertionError('Cannot be both ranged and melee stance')
        if not (ranged_found or melee_found):
            raise AssertionError('Must be ranged or melee stance')
        if ranged_found:
            # ranged damage, and reclaim 2 energy. riptide will attach in
            # skill end.
            return [
                self.charge_self(-3),
                self.attack_opposite_active(
                    match, self.ranged_damage, DamageElementalType.HYDRO),
                self.charge_self(2),
            ]
        else:
            # 7 damage
            return super().get_actions(match)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If ranged stance, call super() to attach riptide.
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        ranged_found: bool = False
        for status in charactor.status:
            if status.name == 'Ranged Stance':
                ranged_found = True
                break
        if not ranged_found:
            # not ranged stance
            return []
        return super().event_handler_SKILL_END(event, match)


class TideWithholder(PassiveSkillBase):
    """
    Only generate Ranged Stance at game start, and generate riptide event
    handler for system. status swicth will be controlled by two status 
    themself.
    """
    name: Literal['Tide Withholder'] = 'Tide Withholder'
    status_name: Literal['Ranged Stance'] = 'Ranged Stance'
    version: Literal['4.1'] = '4.1'

    def event_handler_CHARACTOR_REVIVE(
        self, event: CharactorReviveEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When charactor revive, gain status
        """
        if (
            event.action.player_idx != self.position.player_idx
            or event.action.charactor_idx != self.position.charactor_idx
        ):
            # not this charactor, do nothing
            return []
        return [
            self.create_charactor_status(
                'Ranged Stance',
                { 'version': self.version }
            ),
        ]

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When game begin, gain ranged stance and system handler
        """
        return [
            self.create_charactor_status(
                'Ranged Stance',
                { 'version': self.version }
            ),
        ]


# Talents


class AbyssalMayhemHydrospout_4_1(SkillTalent):
    name: Literal['Abyssal Mayhem: Hydrospout']
    version: Literal['4.1'] = '4.1'
    charactor_name: Literal['Tartaglia'] = 'Tartaglia'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
    )
    skill: Literal['Foul Legacy: Raging Tide'] = 'Foul Legacy: Raging Tide'
    only_active_charactor: bool = True

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        Make 1 damage to character if affected by Riptide. When only active
        charactor (after 4.1), only check active charactor. Otherwise,
        check all charactors.
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        ret: List[MakeDamageAction] = []
        active_idx = match.player_tables[
            1 - self.position.player_idx].active_charactor_idx
        for cidx, charactor in enumerate(
            match.player_tables[1 - self.position.player_idx].charactors
        ):
            if self.only_active_charactor and cidx != active_idx:
                continue
            for status in charactor.status:
                if status.name == 'Riptide':
                    ret.append(MakeDamageAction(
                        damage_value_list = [
                            DamageValue(
                                position = self.position,
                                damage_type = DamageType.DAMAGE,
                                target_position = charactor.position,
                                damage = 1,
                                damage_elemental_type 
                                = DamageElementalType.PIERCING,
                                cost = Cost(),
                            )
                        ]
                    ))
                    break
        return ret


# charactor base


class Tartaglia_4_1(CharactorBase):
    name: Literal['Tartaglia']
    desc: Literal['', 'transform'] = ''
    version: Literal['4.1'] = '4.1'
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | FoulLegacyRagingTide
        | HavocObliteration | TideWithholder
    ] = []
    faction: List[FactionType] = [
        FactionType.FATUI
    ]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Cutting Torrent',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            FoulLegacyRagingTide(),
            HavocObliteration(),
            TideWithholder(),
        ]

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[Actions]:
        status = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx].status
        for s in status:
            if s.name == 'Melee Stance':
                self.desc = 'transform'
                return []
        self.desc = ''
        return []


register_class(Tartaglia_4_1 | AbyssalMayhemHydrospout_4_1)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import (
    AfterMakeDamageEventArguments, CharactorDefeatedEventArguments, 
)

from ...action import CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    CreateStatusPassiveSkill, ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class UpaShato(ElementalBurstBase):
    name: Literal['Upa Shato'] = 'Upa Shato'
    damage: int = 5
    damage_type: DamageElementalType = DamageElementalType.PHYSICAL
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3,
        charge = 2
    )


class InfusedStonehide(CreateStatusPassiveSkill):
    name: Literal['Infused Stonehide'] = 'Infused Stonehide'
    status_name: Literal['Stonehide'] = 'Stonehide'


# Talents


class StonehideReforged_3_3(SkillTalent):
    name: Literal['Stonehide Reforged']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Stonehide Lawachurl'] = 'Stonehide Lawachurl'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 4,
        charge = 2
    )
    skill: Literal['Upa Shato'] = 'Upa Shato'

    opposite_alive: List[int] = []

    def event_handler_AFTER_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Any
    ):
        """
        record enemy alive status
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped, return
            return []
        # reset alive
        self.opposite_alive = []
        assert len(event.action.damage_value_list) > 0
        # if use skill, source of first should be self
        if not self.position.check_position_valid(
            event.action.damage_value_list[0].position, match,
            player_idx_same = True, charactor_idx_same = True,
            target_area = ObjectPositionType.SKILL
        ):
            # not self use skill, return
            return []
        for cid, charactor in enumerate(match.player_tables[
                1 - self.position.player_idx].charactors):
            self.opposite_alive.append(cid)
        return []

    def event_handler_CHARACTOR_DEFEATED(
        self, event: CharactorDefeatedEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If self is using skill, and enemy defeated, re-attach Stonehide and 
        Stone Force.
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped, return
            return []
        if event.action.player_idx == self.position.player_idx:
            # self defeated, return
            return []
        if event.action.charactor_idx not in self.opposite_alive:
            # defeated charactor not in opposite_alive, return
            return []
        # re-attach Stonehide and Stone Force
        return [
            CreateObjectAction(
                object_name = 'Stonehide',
                object_position = self.position.set_area(
                    ObjectPositionType.CHARACTOR_STATUS
                ),
                object_arguments = {}
            )
        ]


# charactor base


class StonehideLawachurl_3_3(CharactorBase):
    name: Literal['Stonehide Lawachurl']
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.GEO
    max_hp: int = 8
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | ElementalSkillBase | UpaShato 
        | InfusedStonehide
    ] = []
    faction: List[FactionType] = [
        FactionType.MONSTER,
        FactionType.HILICHURL
    ]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Plama Lawa',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name = 'Movo Lawa',
                damage_type = DamageElementalType.PHYSICAL,
                cost = ElementalSkillBase.get_cost(self.element),
            ),
            UpaShato(),
            InfusedStonehide()
        ]


register_class(StonehideLawachurl_3_3 | StonehideReforged_3_3)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....modifiable_values import DamageValue

from ....action import MakeDamageAction

from ....event import SkillEndEventArguments

from ....consts import (
    DamageElementalType, DamageType, ObjectPositionType, WeaponType
)

from ....struct import Cost
from .base import RoundEffectWeaponBase


class AquilaFavonia_3_3(RoundEffectWeaponBase):
    name: Literal['Aquila Favonia']
    cost: Cost = Cost(same_dice_number = 3)
    version: Literal['3.3'] = '3.3'
    weapon_type: WeaponType = WeaponType.SWORD
    max_usage_per_round: int = 2

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self is active and opposite use skill and has usage, heal 1 HP
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = False,
            source_area = ObjectPositionType.CHARACTOR,
            target_area = ObjectPositionType.SKILL
        ):
            # not equipped or not opponent use skill
            return []
        if (
            self.position.charactor_idx != match.player_tables[
                self.position.player_idx].active_charactor_idx
        ):
            # not active
            return []
        if self.usage == 0:
            # no usage
            return []
        self.usage -= 1
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        return [MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            ],
        )]


register_class(AquilaFavonia_3_3)



from ...utils import import_all_modules
from .equipment import WeaponBase, ArtifactBase
from .support import SupportBase, LocationBase, ItemBase, CompanionBase


import_all_modules(__file__, __name__)
__all__ = (
    'WeaponBase', 'ArtifactBase', 'SupportBase', 'LocationBase', 
    'ItemBase', 'CompanionBase'
)



from typing import Any, Literal

from ....utils.class_registry import register_class

from ...consts import IconType, SkillType

from ...modifiable_values import DamageIncreaseValue
from .base import RoundCharactorStatus, ShieldCharactorStatus


class UnmovableMountain_3_5(ShieldCharactorStatus):
    name: Literal['Unmovable Mountain'] = 'Unmovable Mountain'
    version: Literal['3.5'] = '3.5'
    usage: int = 2
    max_usage: int = 2


class VermillionHereafter_4_0(RoundCharactorStatus):
    name: Literal['Vermillion Hereafter']
    version: Literal['4.0'] = '4.0'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not current charactor using normal attack
            return value
        # increase damage
        assert mode == 'REAL'
        value.damage += 1
        return value


register_class(UnmovableMountain_3_5 | VermillionHereafter_4_0)



"""
parse instance based on base class. When an object is registered, we will scan
all its parent class, and register it to the corresponding base class. When we
need to instantiate an object with a dict, with given base class, we will find
corresponding class and version, and instantiate it.
"""


import logging
import types
from typing import (
    Any, Dict, List, Optional, Set, Type, Union, get_type_hints, get_args
)

from ..server.consts import ObjectType
from .desc_registry import DescDictType, desc_exist, update_cost, update_desc


"""
The dict to save all registered classes. Key is the base class, value is a dict
with key as name, value is a dict with version as name, and value as class.
In class version dict, it saves all available version of a card, and is sorted
descendingly.
"""
_class_dict: Dict[Type[Any], Dict[str, Dict[int, Type[Any]]]] = {}
_class_version_dict: Dict[Type[Any], Dict[str, List[int]]] = {}
"""
The set that contains all registered base classes.
"""
_class_set = set()


def _is_union_type(t) -> bool:
    return (
        getattr(t, '__class__', None) == types.UnionType
        or getattr(t, '__origin__', None) is Union 
        or getattr(t, '__origin__', None) is Optional
    )


def _version_to_int(version: str) -> int:
    """
    Convert version string to int. For example, 3.2 will be converted to
    32
    """
    return int(float(version) * 10 + 0.1)


def register_base_class(base_class: Type[Any]):
    """
    Register a base class.
    """
    _class_dict[base_class] = {}
    _class_version_dict[base_class] = {}
    _class_set.add(base_class)


def register_class_one(cls: Type[Any]):
    """
    Register one class. If the class is registered, print a warning. If same
    version, name and base class, raise error. If descs not exist, raise error.
    """
    class_name = cls.__name__
    cls_type_hints = get_type_hints(cls)
    name_hints = cls_type_hints['name']
    version_hints = cls_type_hints['version']
    desc_hints = cls_type_hints['desc']
    if len(get_args(desc_hints)) == 0:
        raise AssertionError(
            f'Class {cls} desc type hint is empty or not a Literal'
        )
    descs = get_args(desc_hints)
    version = version_hints.__args__[0]
    type_hints = cls_type_hints['type']
    if len(type_hints.__args__) != 1:
        raise AssertionError(
            f'Class {cls} type hint is empty or more than 1'
        )
    obj_type = type_hints.__args__[0]
    names = name_hints.__args__
    # check if class name is match
    version_from_name = '.'.join(class_name.split('_')[-2:])
    if version_from_name != version:
        raise AssertionError(
            f'Class {cls} version {version} is not equal to version '
            f'from name {version_from_name}'
        )
    for name in names:
        desc_type = obj_type
        if obj_type == ObjectType.CHARACTOR:
            # for charactor, also check its skill descs
            instance = cls(name = name, version = version)
            for skill in instance.skills:
                skill_type = f'SKILL_{name}_{skill.skill_type}'
                if not desc_exist(skill_type, skill.name, version):
                    raise AssertionError(
                        f'Class {cls} name {name} version {version} skill '
                        f'{skill.name} is not found in descs'
                    )
                # register skill costs
                update_cost(skill_type, skill.name, version, skill.cost)
        if obj_type == ObjectType.TALENT:
            # for talent, its type name should change
            target_charactor_names = cls_type_hints['charactor_name'].__args__
            assert len(target_charactor_names) == 1
            desc_type = f'TALENT_{target_charactor_names[0]}'
        for desc in descs:
            if desc == '':
                desc_name = name
            else:
                desc_name = f'{name}_{desc}'
            if desc == '':
                # empty, no extra desc
                if not desc_exist(desc_type, desc_name, version):
                    raise AssertionError(
                        f'Class {cls} name {name} version {version} type '
                        f'{desc_type} is not found in descs'
                    )
            else:
                if not desc_exist(desc_type, desc_name, version):
                    raise AssertionError(
                        f'Class {cls} name {name}(desc: {desc}) version '
                        f'{version} is not found in descs'
                    )
            if 'cost' in cls_type_hints:
                # has cost, register it
                instance = cls(
                    name = name, version = version, position = {
                        'player_idx': -1, 'area': 'INVALID', 'id': -1
                    }
                )
                update_cost(desc_type, desc_name, version, instance.cost)
    version = _version_to_int(version)
    register_success = False
    base_classes = cls.mro()
    for base_class in base_classes:
        if base_class in _class_set:
            if base_class.__name__ == 'CharactorBase':
                # is a charactor, check if talent type hint is original
                talent_hint = cls_type_hints['talent']
                assert len(talent_hint.__args__) == 2
                assert talent_hint.__args__[0].__name__ == 'TalentBase', (
                    cls
                )
                assert talent_hint.__args__[1] is type(None)
            register_success = True
            for name in names:
                if name not in _class_dict[base_class]:
                    _class_dict[base_class][name] = {}
                    _class_version_dict[base_class][name] = []
                target_dict = _class_dict[base_class][name]
                target_version_dict = _class_version_dict[base_class][name]
                if version in target_dict:
                    if target_dict[version] == cls:
                        logging.warning(
                            f'Class {cls} is already registered in base class '
                            f'{base_class.__name__} registry'
                        )
                        continue
                    else:
                        raise AssertionError(
                            f'Class {cls} and {target_dict[version]} have '
                            'same version and name'
                        )
                target_dict[version] = cls
                target_version_dict.append(version)
                target_version_dict.sort(reverse = True)
    if not register_success:
        raise AssertionError(
            f'Class {cls} is not registered in class registry, '
            'No base class is matched.'
        )


def register_class(
        classes: Type[Any], descs: Dict[str, DescDictType] | None = None
):
    """
    Register classes with their descriptions. If classes is a Union, register
    all classes in the union. Otherwise, register the class itself.
    Before registering, descs will be updated first.
    """
    if descs is not None:
        update_desc(descs)
    if _is_union_type(classes):
        for type in classes.__args__:  # type: ignore
            register_class_one(type)
    else:
        register_class_one(classes)


def _get_class(
    class_dict: Dict[str, Dict[int, Type[Any]]], 
    class_version_list: Dict[str, List[int]], 
    name: str, version: str
):
    version_int = _version_to_int(version)
    for v in class_version_list[name]:
        if v <= version_int:
            return class_dict[name][v]
    else:
        # all failed, raise error
        raise KeyError()


def _parse_object(
    class_dict: Dict[str, Dict[int, Type[Any]]], 
    class_version_list: Dict[str, List[int]], 
    name: str, version: str, args: Any
):
    cls = _get_class(class_dict, class_version_list, name, version)
    return cls(**args)


# def get_instance_from_type_unions(types, args, key = 'name'):
def get_instance(base_class: Type[Any], args: Dict):
    """
    Get instance from registry. If the base class is a union type, we will try
    each type sequentially.
    """
    name = args['name']
    version = args.get('version', '99.9')
    if _is_union_type(base_class):
        # is union type, try each class sequentially
        for type in base_class.__args__:  # type: ignore
            if type not in _class_dict:
                raise AssertionError(
                    f'Base class {type} is not registered in class registry'
                )
            try:
                return _parse_object(
                    _class_dict[type], _class_version_dict[type],
                    name, version, args
                )
            except KeyError:
                continue
        raise AssertionError(
            f'Cannot find class {name} in union type {base_class}'
        )
    else:
        if base_class not in _class_dict:
            raise AssertionError(
                f'Base class {base_class} is not registered in class registry'
            )
        try:
            return _parse_object(
                _class_dict[base_class], 
                _class_version_dict[base_class], 
                name,
                version,
                args
            )
        except KeyError:
            raise AssertionError(
                f'Cannot find class {name} with version {version} in base '
                f'class {base_class}'
            )


def get_class_list_by_base_class(
    base_class: Type[Any], version: str = '99.9', exclude: Set[str] = set()
) -> List[str]:
    """
    Get list of class names by base class and version. If the base class is a 
    union type, class names that matchs any type in the union will be returned.

    Args:
        base_class: The base class to search.
        version: The version of the class. Classes that have higher version
            than this will be ignored. Default is '99.9'.
        exclude: Exclude class with specified names in the list.

    Return:
        List of class names.
    """
    result_set: Set[str] = set()
    base_class_list = [base_class]
    if _is_union_type(base_class):
        # is union type, try each class sequentially
        base_class_list = base_class.__args__  # type: ignore
    for type in base_class_list:
        if type not in _class_dict:
            raise AssertionError(
                f'Base class {type} is not registered in class registry'
            )
        names = _class_dict[type].keys()
        for name in names:
            if name in exclude:
                continue
            try:
                _ = _get_class(
                    _class_dict[type], 
                    _class_version_dict[type], 
                    name, 
                    version
                )
                # get class success
                result_set.add(name)
            except KeyError:
                continue
    return sorted(list(result_set))


__all__ = (
    'register_base_class', 
    'register_class', 
    'get_instance',
    'get_class_list_by_base_class',
)



from typing import List, Literal

from ....utils.class_registry import register_class

from ..charactor_base import PhysicalNormalAttackBase
from .shenhe_3_7 import (
    Shenhe_3_7, SpringSpiritSummoning as SS_3_7, DivineMaidensDeliverance
)


class SpringSpiritSummoning(SS_3_7):
    max_usage: int = 2


class Shenhe_4_2(Shenhe_3_7):
    version: Literal['4.2'] = '4.2'
    skills: List[
        PhysicalNormalAttackBase | SpringSpiritSummoning 
        | DivineMaidensDeliverance
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Dawnstar Piercer',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SpringSpiritSummoning(),
            DivineMaidensDeliverance(),
        ]


register_class(Shenhe_4_2)



from ...utils import import_all_modules
from .charactor_base import CharactorBase, TalentBase, SkillBase


import_all_modules(__file__, __name__, exceptions = {'template'})
__all__ = ('CharactorBase', 'TalentBase', 'SkillBase')
